import unittest

from pathlib import Path
import numpy as np
import sympy as sp
import logging

from shgpy.fformgen import (
    _save_fform_dict,
    _load_fform_dict,
)

_logger = logging.getLogger(__name__)


class TestPickle(unittest.TestCase):

    def test_pickle(self):
        path = Path(__file__).parent / 'fform' / 'test_pickle_fform.p'
        M = 16
        x = sp.symbols('x', real=True)
        iterable = {}
        for k in ['PP', 'PS', 'SP', 'SS']:
            iterable[k] = np.zeros(2*M+1, dtype=sp.Expr)
            for i in np.arange(-16, 17):
                iterable[k][i] = x**2

        _save_fform_dict(path, iterable)
        fform_dict = _load_fform_dict(path)
        path.unlink()
        
        self.assertNotEqual(fform_dict['SS'][0].subs(sp.symbols('x'), 1), 1)
        self.assertEqual(fform_dict['SS'][0].subs(sp.symbols('x', real=True), 1), 1)
        self.assertEqual(fform_dict['SS'][0], x**2)



"""Fourier fitting module for fitting RA-SHG data in Fourier space.

This module provides two different methods (up to minor variations) for
fitting RA-SHG data in Fourier space. The first is a simple
least-square method (using `scipy.optimize.least_squares`) which is
extremely fast and works well for simple problems. The second is
a so-called basinhopping method (using `scipy.optimize.basinhopping`)
which is proficient at finding the global minimum of complicated cost
functions with many fitting parameters and local minima. See the
relevant `scipy` documentation for more info.

"""
import sympy as sp
import numpy as np
from .core import n2i
from scipy.optimize import (
    basinhopping,
    least_squares,
    OptimizeResult,
)
import time
import logging
from warnings import warn

_logger = logging.getLogger(__name__)


def _check_fform(fform):
    if not fform.get_free_symbols():
        message = 'fFormContainer object has no free symbols to use as fitting parameters. Is your fitting formula actually zero?'
        warn(message)
        ret = OptimizeResult(
            x=np.array([]),
            xdict={},
            success=False,
            status=0,
            message=message,
        )
        return ret


def _I_component(expr):
    return (expr-expr.subs(sp.I, 0)).subs(sp.I, 1)


def _no_I_component(expr):
    return expr.subs(sp.I, 0)


def least_squares_fit(fform, fdat, guess_dict):
    """Least-squares fit of RA-SHG data.

    Parameters
    ----------
    fform : fFormContainer
        Instance of class :class:`~shgpy.core.data_handler.fFormContainer`.
        This is the (Fourier-transformed) fitting formula.
    fdat : fDataContainer
        Instance of class :class:`~shgpy.core.data_handler.fDataContainer`.
        This is the (Fourier-transformed) data to fit.
    guess_dict : dict
        Dict of form ``{sympy.Symbol:float}``. This is the initial guess.

    Returns
    -------
    ret : scipy.optimize.OptimizeResult
        Instance of class :class:`~scipy.optimize.OptimizeResult`.
        See `scipy` documentation for further description. Includes
        additional attribute ``ret.xdict`` which is a `dict` of 
        ``{sympy.Symbol:float}`` indicating the final answer as
        a dictionary.

    """
    free_symbols = fform.get_free_symbols()
    M = fform.get_M()

    expr_residual_list = []
    for pc in fform.get_keys():
        for m in np.arange(-M, M+1):
            expr_residual_list.append(fform.get_pc(pc)[n2i(m, M)] - fdat.get_pc(pc)[n2i(m, M)])

    _logger.info('Starting residual function generation.')
    start = time.time()
    pre_residual = sp.lambdify(free_symbols, expr_residual_list)
    residual = lambda x:np.array(pre_residual(*x)).view(np.double)
    _logger.info(f'Done with residual function generation. It took {time.time()-start} seconds.')

    guess = [guess_dict[k] for k in free_symbols]

    _logger.info('Starting least squares minimizations.')
    start = time.time()
    ret = least_squares(residual, guess)
    ret.time = time.time()-start
    ret.xdict = {k:ret.x[i] for i,k in enumerate(free_symbols)}
    _logger.info(f'Finished least squares minimization. It took {time.time()-start} seconds.')
    
    return ret


def least_squares_fit_with_bounds(fform, fdat, guess_dict, bounds_dict):
    """Least-squares fit of RA-SHG data with bounds.

    Parameters
    ----------
    fform : fFormContainer
        Instance of class :class:`~shgpy.core.data_handler.fFormContainer`.
        This is the (Fourier-transformed) fitting formula.
    fdat : fDataContainer
        Instance of class :class:`~shgpy.core.data_handler.fDataContainer`.
        This is the (Fourier-transformed) data to fit.
    guess_dict : dict
        Dict of form ``{sympy.Symbol:float}``. This is the initial guess.
    bounds_dict : dict
        Dict of form ``{sympy.Symbol:tuple}``. `tuple` should be of form
        ``(lower_bound, upper_bound)``.

    Returns
    -------
    ret : scipy.optimize.OptimizeResult
        Instance of class :class:`~scipy.optimize.OptimizeResult`.
        See `scipy` documentation for further description. Includes
        additional attribute ``ret.xdict`` which is a `dict` of 
        ``{sympy.Symbol:float}`` indicating the final answer as
        a dictionary.

    """
    check = _check_fform(fform)
    if check:
        return check
    free_symbols = fform.get_free_symbols()
    M = fform.get_M()

    expr_residual_list = []
    for pc in fform.get_keys():
        for m in np.arange(-M, M+1):
            expr_residual_list.append(fform.get_pc(pc)[n2i(m, M)] - fdat.get_pc(pc)[n2i(m, M)])

    _logger.info('Starting residual function generation.')
    start = time.time()
    pre_residual = sp.lambdify(free_symbols, expr_residual_list)
    residual = lambda x:np.array(pre_residual(*x)).view(np.double)
    _logger.info(f'Done with residual function generation. It took {time.time()-start} seconds.')

    guess = [guess_dict[k] for k in free_symbols]
    bounds = [[bounds_dict[k][0] for k in free_symbols], [bounds_dict[k][1] for k in free_symbols]]

    _logger.info('Starting least squares minimizations.')
    start = time.time()
    ret = least_squares(residual, guess, bounds=bounds)
    ret.time = time.time()-start
    ret.xdict = {k:ret.x[i] for i,k in enumerate(free_symbols)}
    _logger.info(f'Finished least squares minimization. It took {time.time()-start} seconds.')
    
    return ret


def basinhopping_fit(fform, fdat, guess_dict, niter, method='BFGS', stepsize=0.5):
    """Basinhopping fit of RA-SHG data.

    Parameters
    ----------
    fform : fFormContainer
        Instance of class :class:`~shgpy.core.data_handler.fFormContainer`.
        This is the (Fourier-transformed) fitting formula.
    fdat : fDataContainer
        Instance of class :class:`~shgpy.core.data_handler.fDataContainer`.
        This is the (Fourier-transformed) data to fit.
    guess_dict : dict
        Dict of form ``{sympy.Symbol:float}``. This is the initial guess.
    niter : int
        Number of basinhopping iterations (see scipy documentation)
    method : str, optional
        Minimization method to use, defaults to `'BFGS'`. See scipy
        documentation for more information.
    stepsize : float, optional
        Basinhopping stepsize, defaults to `0.5`. See scipy documentation
        for more information.

    Returns
    -------
    ret : scipy.optimize.OptimizeResult
        Instance of class :class:`~scipy.optimize.OptimizeResult`.
        See `scipy` documentation for further description. Includes
        additional attribute ``ret.xdict`` which is a `dict` of 
        ``{sympy.Symbol:float}`` indicating the final answer as
        a dictionary.

    """
    check = _check_fform(fform)
    if check:
        return check
    free_symbols = fform.get_free_symbols()
    M = fform.get_M()

    _logger.info('Starting energy function generation.')
    start = time.time()
    energy_expr_list = []
    for k in fform.get_keys():
        for m in np.arange(-M, M+1):
            expr1 = _no_I_component(fform.get_pc(k)[n2i(m, M)]-fdat.get_pc(k)[n2i(m, M)])
            expr2 = _I_component(fform.get_pc(k)[n2i(m, M)]-fdat.get_pc(k)[n2i(m, M)])
            energy_expr_list.append(expr1**2)
            energy_expr_list.append(expr2**2)

    energy_expr = sum(energy_expr_list)

    pre_f_energy = sp.lambdify(free_symbols, energy_expr)
    f_energy = lambda x:pre_f_energy(*x)
    _logger.info(f'Done with energy function generation. It took {time.time()-start} seconds.')

    x0 = [guess_dict[k] for k in free_symbols]
    minimizer_kwargs = {'method':method}
    _logger.info('Starting basinhopping minimization.')
    start = time.time()
    ret = basinhopping(f_energy, x0, minimizer_kwargs=minimizer_kwargs, niter=niter, stepsize=stepsize)
    ret.time = time.time()-start
    ret.xdict = {k:ret.x[i] for i,k in enumerate(free_symbols)}
    _logger.info(f'Done with basinhopping minimization. It took {ret.time} seconds.')

    return ret


def basinhopping_fit_with_bounds(fform, fdat, guess_dict, bounds_dict, niter, method='L-BFGS-B', stepsize=0.5):
    """Basinhopping fit of RA-SHG data with bounds.

    Parameters
    ----------
    fform : fFormContainer
        Instance of class :class:`~shgpy.core.data_handler.fFormContainer`.
        This is the (Fourier-transformed) fitting formula.
    fdat : fDataContainer
        Instance of class :class:`~shgpy.core.data_handler.fDataContainer`.
        This is the (Fourier-transformed) data to fit.
    guess_dict : dict
        Dict of form ``{sympy.Symbol:float}``. This is the initial guess.
    bounds_dict : dict
        Dict of form ``{sympy.Symbol:tuple}``. `tuple` should be of form
        ``(lower_bound, upper_bound)``.
    niter : int
        Number of basinhopping iterations (see scipy documentation)
    method : str, optional
        Minimization method to use, defaults to `'L-BFGS-B'`. See scipy
        documentation for more information.
    stepsize : float, optional
        Basinhopping stepsize, defaults to `0.5`. See scipy documentation
        for more information.

    Returns
    -------
    ret : scipy.optimize.OptimizeResult
        Instance of class :class:`~scipy.optimize.OptimizeResult`.
        See `scipy` documentation for further description. Includes
        additional attribute ``ret.xdict`` which is a `dict` of 
        ``{sympy.Symbol:float}`` indicating the final answer as
        a dictionary.

    """
    check = _check_fform(fform)
    if check:
        return check
    free_symbols = fform.get_free_symbols()
    M = fform.get_M()

    _logger.info('Starting energy function generation.')
    start = time.time()
    energy_expr_list = []
    for k in fform.get_keys():
        for m in np.arange(-M, M+1):
            expr1 = _no_I_component(fform.get_pc(k)[n2i(m, M)]-fdat.get_pc(k)[n2i(m, M)])
            expr2 = _I_component(fform.get_pc(k)[n2i(m, M)]-fdat.get_pc(k)[n2i(m, M)])
            energy_expr_list.append(expr1**2)
            energy_expr_list.append(expr2**2)

    energy_expr = sum(energy_expr_list)

    pre_f_energy = sp.lambdify(free_symbols, energy_expr)
    f_energy = lambda x:pre_f_energy(*x)
    _logger.info(f'Done with energy function generation. It took {time.time()-start} seconds.')

    x0 = [guess_dict[k] for k in free_symbols]
    if bounds_dict is not None:
        bounds = [bounds_dict[k] for k in free_symbols]
    if bounds_dict is None:
        bounds = None
    minimizer_kwargs = {'method':method, 'bounds':bounds}
    start = time.time()
    _logger.info('Starting basinhopping minimization.')
    ret = basinhopping(f_energy, x0, minimizer_kwargs=minimizer_kwargs, niter=niter, stepsize=stepsize)
    ret.time = time.time()-start
    ret.xdict = {k:ret.x[i] for i,k in enumerate(free_symbols)}
    _logger.info(f'Done with basinhopping minimization. It took {ret.time} seconds.')

    return ret


def basinhopping_fit_jac(fform, fdat, guess_dict, niter, method='BFGS', stepsize=0.5):
    """Basinhopping fit of RA-SHG data.

    Parameters
    ----------
    fform : fFormContainer
        Instance of class :class:`~shgpy.core.data_handler.fFormContainer`.
        This is the (Fourier-transformed) fitting formula.
    fdat : fDataContainer
        Instance of class :class:`~shgpy.core.data_handler.fDataContainer`.
        This is the (Fourier-transformed) data to fit.
    guess_dict : dict
        Dict of form ``{sympy.Symbol:float}``. This is the initial guess.
    niter : int
        Number of basinhopping iterations (see scipy documentation)
    method : str, optional
        Minimization method to use, defaults to `'BFGS'`. See scipy
        documentation for more information.
    stepsize : float, optional
        Basinhopping stepsize, defaults to `0.5`. See scipy documentation
        for more information.

    Returns
    -------
    ret : scipy.optimize.OptimizeResult
        Instance of class :class:`~scipy.optimize.OptimizeResult`.
        See `scipy` documentation for further description. Includes
        additional attribute ``ret.xdict`` which is a `dict` of 
        ``{sympy.Symbol:float}`` indicating the final answer as
        a dictionary.

    Notes
    -----
    This function computes and supplies the gradient function to the scipy
    basinhopping algorithm. This requires some computational power up front
    but can speed up the minimization algorithm.

    """
    check = _check_fform(fform)
    if check:
        return check
    free_symbols = fform.get_free_symbols()
    M = fform.get_M()

    _logger.info('Starting energy function generation.')
    start = time.time()
    energy_expr_list = []
    for k in fform.get_keys():
        for m in np.arange(-M, M+1):
            expr1 = _no_I_component(fform.get_pc(k)[n2i(m, M)]-fdat.get_pc(k)[n2i(m, M)])
            expr2 = _I_component(fform.get_pc(k)[n2i(m, M)]-fdat.get_pc(k)[n2i(m, M)])
            energy_expr_list.append(expr1**2)
            energy_expr_list.append(expr2**2)

    def gradient(expr, free_symbols):
        return np.array([sp.diff(expr, fs) for fs in free_symbols])

    energy_expr = sum(energy_expr_list)
    denergy_expr = gradient(energy_expr, free_symbols)

    pre_f_energy = sp.lambdify(free_symbols, energy_expr)
    pre_df_energy = sp.lambdify(free_symbols, denergy_expr)
    fdf_energy = lambda x:(pre_f_energy(*x),np.array(pre_df_energy(*x)))
    _logger.info(f'Done with energy function generation. It took {time.time()-start} seconds.')

    x0 = [guess_dict[k] for k in free_symbols]
    minimizer_kwargs = {'method':method, 'jac':True}
    _logger.info('Starting basinhopping minimization.')
    start = time.time()
    ret = basinhopping(fdf_energy, x0, minimizer_kwargs=minimizer_kwargs, niter=niter, stepsize=stepsize)
    ret.time = time.time()-start
    ret.xdict = {k:ret.x[i] for i,k in enumerate(free_symbols)}
    _logger.info(f'Done with basinhopping minimization. It took {ret.time} seconds.')

    return ret


def basinhopping_fit_jac_with_bounds(fform, fdat, guess_dict, bounds_dict, niter, method='L-BFGS-B', stepsize=0.5):
    """Basinhopping fit of RA-SHG data with bounds.

    Parameters
    ----------
    fform : fFormContainer
        Instance of class :class:`~shgpy.core.data_handler.fFormContainer`.
        This is the (Fourier-transformed) fitting formula.
    fdat : fDataContainer
        Instance of class :class:`~shgpy.core.data_handler.fDataContainer`.
        This is the (Fourier-transformed) data to fit.
    guess_dict : dict
        Dict of form ``{sympy.Symbol:float}``. This is the initial guess.
    bounds_dict : dict
        Dict of form ``{sympy.Symbol:tuple}``. `tuple` should be of form
        ``(lower_bound, upper_bound)``.
    niter : int
        Number of basinhopping iterations (see scipy documentation)
    method : str, optional
        Minimization method to use, defaults to `'L-BFGS-B'`. See scipy
        documentation for more information.
    stepsize : float, optional
        Basinhopping stepsize, defaults to `0.5`. See scipy documentation
        for more information.

    Returns
    -------
    ret : scipy.optimize.OptimizeResult
        Instance of class :class:`~scipy.optimize.OptimizeResult`.
        See `scipy` documentation for further description. Includes
        additional attribute ``ret.xdict`` which is a `dict` of 
        ``{sympy.Symbol:float}`` indicating the final answer as
        a dictionary.

    Notes
    -----
    This function computes and supplies the gradient function to the scipy
    basinhopping algorithm. This requires some computational power up front
    but can speed up the minimization algorithm.

    """
    check = _check_fform(fform)
    if check:
        return check
    free_symbols = fform.get_free_symbols()
    M = fform.get_M()

    _logger.info('Starting energy function generation.')
    start = time.time()
    energy_expr_list = []
    for k in fform.get_keys():
        for m in np.arange(-M, M+1):
            expr1 = _no_I_component(fform.get_pc(k)[n2i(m, M)]-fdat.get_pc(k)[n2i(m, M)])
            expr2 = _I_component(fform.get_pc(k)[n2i(m, M)]-fdat.get_pc(k)[n2i(m, M)])
            energy_expr_list.append(expr1**2)
            energy_expr_list.append(expr2**2)

    def gradient(expr, free_symbols):
        return np.array([sp.diff(expr, fs) for fs in free_symbols])

    energy_expr = sum(energy_expr_list)
    denergy_expr = gradient(energy_expr, free_symbols)

    pre_f_energy = sp.lambdify(free_symbols, energy_expr)
    pre_df_energy = sp.lambdify(free_symbols, denergy_expr)
    fdf_energy = lambda x:(pre_f_energy(*x),np.array(pre_df_energy(*x)))
    _logger.info(f'Done with energy function generation. It took {time.time()-start} seconds.')

    x0 = [guess_dict[k] for k in free_symbols]
    if bounds_dict is not None:
        bounds = [bounds_dict[k] for k in free_symbols]
    if bounds_dict is None:
        bounds = None
    minimizer_kwargs = {'method':method, 'jac':True, 'bounds':bounds}
    _logger.info('Starting basinhopping minimization.')
    start = time.time()
    ret = basinhopping(fdf_energy, x0, minimizer_kwargs=minimizer_kwargs, niter=niter, stepsize=stepsize)
    ret.time = time.time()-start
    ret.xdict = {k:ret.x[i] for i,k in enumerate(free_symbols)}
    _logger.info(f'Done with basinhopping minimization. It took {ret.time} seconds.')

    return ret



import numpy as np
import sympy as sp
import csv
import pickle
from scipy.interpolate import interp1d
from scipy.integrate import quad
import logging
import itertools
from copy import deepcopy
from .. import shg_symbols as S
from warnings import warn

_logger = logging.getLogger(__name__)


class DataContainer:
    """Contains SHG data in `phi`-space.

    Parameters
    ----------
    iterable : dict or array_like
        dict which defines the data. Should be in form
        `{'PP':array([xdata, ydata]), 'PS':...}`. Also accepts `array_like`
        of form `(('PP':array([xdata, ydata])), ('PS', ...), ...)`.
    input_angle_units : {'radians', 'degrees'}
        Units of `xdata`.
    normal_to_oblique : bool, optional
        Defaults to ``False``. If ``True``, iterable should have only
        two elements: one with key 'PP' or 'SS', and one with key 'PS'
        or 'SP'. Then, upon initialization, compute the "missing"
        polarization components by rotating by 90 degrees. The final
        object has ``dat.get_keys() = ['PP', 'PS', 'SP', 'SS']``.

    Notes
    -----
    The function of the `DataContainer` class is to hold data which could
    just as well be contained in a dictionary defined like
    ``data = {'PP':array([xdata, ydata]), 'PS':array([xdata, ydata]), ...}``.
    Many of the associated methods involve getting and setting this data like
    you would in a dictionary. However, putting this data in a predefined objects
    allows one to keep track of whatever manipulations (scaling, offsetting) have
    been applied to the data, as well as other attributes like the units with
    which the data was defined.

    """
    def __init__(self, iterable, input_angle_units, normal_to_oblique=False):
        self._input_angle_units = input_angle_units.lower()
        self._check_angle_units(input_angle_units)
        self._check_defining_iterable(iterable)
        self._input_angle_units = input_angle_units
        self._scale = 1
        self._data_dict = deepcopy(dict(iterable))
        self._phase_shift = 0
        self._offset = 0
        if input_angle_units == 'degrees':
            self._convert_data_dict_to_radians()
        if normal_to_oblique:
            self._coerce_to_oblique_incidence()
        self._remove_duplicates()
        self._modulate_data_dict()
        self._sort_data_dict()

    def _check_angle_units(self, angle_units):
        if angle_units not in ['radians', 'degrees']:
            raise ValueError('angle_units input must be one of \'radians\' or \'degrees\'.')

    def _check_defining_iterable(self, iterable):
        try:
            bad_key = False in [type(k) == str for k in dict(iterable).keys()]
            bad_val = False in [type(v) == np.ndarray and v[0].shape == v[1].shape and v.dtype != object and v.ndim == 2 for v in dict(iterable).values()]
        except:
            raise ValueError('Invalid data input')
        if bad_key or bad_val:
            raise ValueError(f'Invalid data input: bad {"key" if bad_key else "value"}')

    def _remove_duplicates(self):
        for k,v in self._data_dict.items():
            new_xdata = []
            new_ydata = []
            for x,y in zip(v[0], v[1]):
                if not (x % (2*np.pi)) in new_xdata:
                    new_xdata.append(x)
                    new_ydata.append(y)
            if len(new_xdata) != len(v[0]):
                warn('Duplicate values encountered in data file.')
            new_xydata = list(zip(*zip(new_xdata, new_ydata)))
            self._data_dict[k] = np.array(new_xydata, dtype=v.dtype)

    def _coerce_to_oblique_incidence(self): 
        keys = set((k.upper()[:2] for k in self._data_dict.keys()))
        if keys in [set(['PP', 'PS', 'SP', 'SS'])]:
            key_type = 'oblique'
        elif keys in [set(i) for i in itertools.product(['PP', 'SS'], ['PS', 'SP'])]:
            key_type = 'normal'
        else:
            raise ValueError("Invalid input: keys must be 'PP', 'PS', 'SP', 'SS' or a nonparallel combination of two.")
        
        if key_type == 'oblique':
            self._data_dict = {k.upper()[:2]:self._data_dict[k] for k in self._data_dict.keys()}
        if key_type == 'normal':
            self._data_dict = {k.upper()[:2]:self._data_dict[k] for k in self._data_dict.keys()}
            leftover = ['PP', 'PS', 'SP', 'SS']
            for k in self._data_dict.keys():
                leftover.remove(k)
            for k in leftover:
                if k[0] == k[1]:
                    copy_pc = {'P':'S', 'S':'P'}[k[0]]*2
                else:
                    copy_pc = k[1]+k[0]
                self._data_dict[k] = deepcopy(self._data_dict[copy_pc])
                for i,x in enumerate(self._data_dict[k][0]):
                    self._data_dict[k][0][i] = (x + np.pi/2) % (2*np.pi)

    def _sort_data_dict(self):
        for k,v in self._data_dict.items():
            new_v = np.array(list(zip(*sorted(zip(v[0], v[1]), key=lambda x:x[0]))))
            self._data_dict[k] = new_v

    def _modulate_data_dict(self):
        for k,v in self._data_dict.items():
            xdata, ydata = v
            new_xdata = xdata % (2*np.pi)
            self._data_dict[k] = np.array([new_xdata, ydata], dtype=v.dtype)
                
    def _convert_data_dict_to_radians(self):
        self._data_dict = {k:np.array([np.deg2rad(v[0]), np.copy(v[1])]) for k,v in self._data_dict.items()}

    def _get_data_dict_degrees(self):
        return {k:np.array([np.rad2deg(v[0]), np.copy(v[1])]) for k,v in self._data_dict.items()}

    def scale_data(self, scale_factor):
        """Scale the data by a constant factor."""
        if scale_factor < 0:
            raise ValueError('scale factor cannot be negative.')
        self._data_dict = {k:np.array([v[0], scale_factor*v[1]]) for k,v in self._data_dict.items()}
        self._scale *= scale_factor

    def subtract_min(self):
        """Subtract the minimum from the data."""
        minval = self.get_minval()[1]
        for k,v in self.get_items('radians'):
            self._data_dict[k] = np.array([v[0], v[1]-minval])
        self._offset -= minval

    def get_maxval(self):
        """Get the maximum value of the data.
        
        Returns
        -------
        mac_pc : str
            The polarization combination for which the maximum occurred
        maximum_value : float
            The maximum value attained.
        """
        maximum_value = -np.inf
        max_pc = None
        for k,v in self._data_dict.items():
            if max(v[1]) > maximum_value:
                maximum_value = max(v[1])
                max_pc = k
        return max_pc, maximum_value

    def get_pc_maxval(self, pc):
        """Get the maximum value of a particular polarization combination.
        
        Parameters
        ----------
        pc : str
            The polarization combination of interest.
        
        Returns
        -------
        maximum_value : float
        """
        return max(self._data_dict[pc][1])

    def get_minval(self):
        """Get the minimum value of the data.

        Returns
        -------
        min_pc : str
            The polarization combination for which the minimum occurred.
        minimum_value : float
            The minimum value attained.
        """
        minimum_value = np.inf
        min_pc = None
        for k,v in self._data_dict.items():
            if min(v[1]) < minimum_value:
                minimum_value = min(v[1])
                min_pc = k
        return min_pc, minimum_value

    def get_pc_minval(self, pc):
        """Get the minimum value of a particular polarization combination.
        
        Parameters
        ----------
        pc : str
            The polarization combination of interest.
        
        Returns
        -------
        minimum_value : float
        """
        return min(self._data_dict[pc][1])

    def get_offset(self):
        """Returns the degree to which the data has been offset"""
        return self._offset

    def normalize_data(self, desired_maximum=1):
        """Normalize the data to a desired maximum."""
        _, maximum_value = self.get_maxval()
        self.scale_data(desired_maximum / maximum_value)

    def phase_shift_data(self, angle, angle_units):
        """Phase shift the data by a constant amount.
        
        Parameters
        ----------
        angle : float
        angle_units : {'radians', 'degrees'}
            Units of argument `angle`.
        """
        angle_units = angle_units.lower()
        self._check_angle_units(angle_units)
        if angle_units == 'degrees':
            angle = np.deg2rad(angle)
        self._data_dict = {k:np.array([(v[0]+angle) % (2*np.pi), v[1]]) for k,v in self._data_dict.items()}
        self._phase_shift += angle
        self._sort_data_dict()

    def get_pc(self, pc, requested_angle_units):
        """Get data for single polarization combination.

        Parameters
        ----------
        pc : str
            Polarization combination of interest.
        requested_angle_units : {'radians', 'degrees'}
            Units requested for `xdata`.

        Returns
        -------
        data : ndarray
            Data for polarization combination `pc` in the form
            `array([xdata, ydata])`.
        """
        requested_angle_units = requested_angle_units.lower()
        self._check_angle_units(requested_angle_units)
        if requested_angle_units == 'radians':
            return np.copy(self._data_dict[pc])
        if requested_angle_units == 'degrees':
            return np.copy(self._get_data_dict_degrees()[pc])

    def get_keys(self):
        """Get the polarization combinations for this data.

        Equivalent to `dict.keys()`.
        
        Returns
        -------
        pcs : list of str
        """
        return list(self._data_dict.keys())

    def get_values(self, requested_angle_units):
        """Get a list of data (without polarization combination reference)

        Equivalent to `dict.values()`.

        Parameters
        ----------
        requested_angle_units : {'radians', 'degrees'}
            Units requested for `xdata`.

        Returns
        -------
        data : list of ndarray
            Data for each polarization combination
            in the form `array([xdata, ydata])`
        """
        requested_angle_units = requested_angle_units.lower()
        self._check_angle_units(requested_angle_units)
        if requested_angle_units == 'radians':
            return list(self._data_dict.values())
        elif requested_angle_units == 'degrees':
            return list(self._get_data_dict_degrees().values())

    def get_items(self, requested_angle_units):
        """Get a list of `(pc, data)` pairs.

        Equivalent to `dict.items()`.

        Parameters
        ----------
        requested_angle_units : {'radians', 'degrees'}
            Units requested for `xdata`.

        Returns
        -------
        data : list of (pc, ndarray) tuple
            List of data for each polarization combination, each
            in the form `(pc, array([xdata, ydata]))`
        """
        requested_angle_units = requested_angle_units.lower()
        self._check_angle_units(requested_angle_units)
        if requested_angle_units == 'radians':
            return list(self._data_dict.items())
        elif requested_angle_units == 'degrees':
            return list(self._get_data_dict_degrees().items())

    def get_scale(self):
        """Get the extent to which this data has been scaled from the original."""
        return self._scale

    def get_phase_shift(self, requested_angle_units):
        """Get the extent to which this data has been phase shifted from the original.

        Parameters
        ----------
        requested_angle_units : {'radians', 'degrees'}

        Returns
        -------
        phase_shift : float
        """
        requested_angle_units = requested_angle_units.lower()
        self._check_angle_units(requested_angle_units)
        if requested_angle_units == 'radians':
            return self._phase_shift
        elif requested_angle_units == 'degrees':
            return np.rad2deg(self._phase_shift)


class fDataContainer:
    """Contains SHG data in Fourier space.

    Parameters
    ----------
    iterable : dict or array_like
        dict which defines the data. Should be in form
        `{'PP':ndarray(...), 'PS':..., ...}`. Also accepts `array_like`
        of form `(('PP':ndarray(...)), ('PS', ...), ...)`.

    Notes
    -----
    A central capability of the `shgpy` package is to be able to fit
    RA-SHG data in Fourier space. What that amounts to is fitting the
    Fourier components of a particular fitting formula to the Fourier
    components of the data. An instance of the `fDataContainer` class
    basically acts as a dictionary holding pairs `(pc, fdata)`, where
    `pc` is a `str` label for the polarization combination and `fdata`
    is an ndarray of length `2*M+1`. The `nth` element here denotes the
    `(n-M)th` fourier component of the data to be fitted. 
    """

    def __init__(self, iterable, M=16):
        self._M = M
        self._check_defining_iterable(iterable)
        self._fdata_dict = deepcopy(dict(iterable))
        self._scale = 1
        self._phase_shift = 0

    def _check_defining_iterable(self, iterable):
        def raise_error():
            raise ValueError('Invalid Data input')
        try:
            bad_key = False in [type(k) == str for k in dict(iterable).keys()]
            bad_val = False in [type(v) == np.ndarray and v.shape == (2*self._M+1,) and v.dtype in [np.complex64, np.complex128] and v.ndim == 1 for v in dict(iterable).values()]
        except:
            raise_error()
        if bad_key or bad_val:
            raise_error()

    def _check_angle_units(self, angle_units):
        if angle_units not in ['radians', 'degrees']:
            raise ValueError('angle_units input must be one of \'radians\' or \'degrees\'.')

    def get_keys(self):
        """Get the polarization combinations for this data.

        Equivalent to `dict.keys()`.
        
        Returns
        -------
        pcs : list of str
        """
        return list(self._fdata_dict.keys())

    def get_values(self):
        """Get a list of data (without polarization combination reference)

        Equivalent to `dict.values()`.

        Returns
        -------
        fdata : list of ndarray
            Data for each polarization combination
            in the form `array(shape=(2*M+1,))`
        """
        return list(self._fdata_dict.values())

    def get_items(self):
        """Get a list of `(pc, fdata)` pairs.

        Equivalent to `dict.items()`.

        Returns
        -------
        fdata : list of (pc, ndarray) tuple
            List of data for each polarization combination, each
            in the form `(pc, array(shape=(2*M+1,)))`
        """
        return list(self._fdata_dict.items())

    def get_scale(self):
        """Get the extent to which this data has been scaled from the original."""
        return self._scale

    def get_M(self):
        """Get the number of Fourier frequencies.

        Each `fdata` array has `2*M+1` elements."""
        return self._M

    def get_phase_shift(self, requested_angle_units):
        """Get the extent to which this data has been phase shifted from the original.

        Parameters
        ----------
        requested_angle_units : {'radians', 'degrees'}

        Returns
        -------
        phase_shift : float
        """
        requested_angle_units = requested_angle_units.lower()
        self._check_angle_units(requested_angle_units)
        if requested_angle_units == 'degrees':
            return np.rad2deg(self._phase_shift)
        return self._phase_shift
        
    def scale_fdata(self, scale_factor):
        """Normalize the fdata by a given amount."""
        if scale_factor < 0:
            raise ValueError('scale factor cannot be negative.')
        self._fdata_dict = {k:scale_factor*v for k,v in self.get_items()}
        self._scale *= scale_factor

    def get_pc(self, pc):
        """Get fdata for single polarization combination.

        Parameters
        ----------
        pc : str
            Polarization combination of interest.

        Returns
        -------
        fdata : ndarray
            Data for polarization combination `pc` with shape `2*M+1`. 
        """
        return np.copy(self._fdata_dict[pc])

    def get_maxval(self):
        """Get the maximum value of the data.
        
        Returns
        -------
        mac_pc : str
            The polarization combination for which the maximum occurred
        maximum_value : float
            The maximum value attained.

        Notes
        -----
        Since the elements of the fourier transform are in general
        complex numbers, this method returns the maximum value of
        the magnitudes of the different elements.
        """
        maximum_value = -np.inf
        max_pc = None
        for k,v in self.get_items():
            this_max = max([abs(nv) for nv in v])
            if this_max > maximum_value:
                maximum_value = this_max
                max_pc = k
        return max_pc, maximum_value

    def get_pc_maxval(self, pc):
        """Get the maximum value of a particular polarization combination.
        
        Parameters
        ----------
        pc : str
            The polarization combination of interest.
        
        Returns
        -------
        maximum_value : float
            Returns maximum of magnitudes of fourier components
            (see :func:`~shgpy.core.data_handler.fDataContainer.get_maxval`)
        """
        v = self._fdata_dict[pc]
        maximum_value = max([abs(nv) for nv in v])
        return maximum_value
        
    def normalize_fdata(self, desired_maximum):
        """Normalize the data to a desired maximum."""
        _, maximum_value = self.get_maxval()
        self.scale_fdata(desired_maximum / maximum_value)

    def phase_shift_fdata(self, angle, angle_units):
        """Phase shift the data by a constant amount.
        
        Parameters
        ----------
        angle : float
        angle_units : {'radians', 'degrees'}
            Units of argument `angle`.
        """
        angle_units = angle_units.lower()
        self._check_angle_units(angle_units)
        if angle_units == 'degrees':
            angle = np.deg2rad(angle)
        for k in self.get_keys():
            fexpr = self._fdata_dict[k]
            ans = np.zeros(shape=fexpr.shape, dtype=fexpr.dtype)
            for m in np.arange(-self._M, self._M + 1):
                ans[n2i(m, self._M)] = fexpr[n2i(m, self._M)] * (np.cos(m * angle)+1j*np.sin(m * angle))
            self._fdata_dict[k] = ans
        self._phase_shift += angle


class fFormContainer:
    """Contains an SHG formula in Fourier space.

    Parameters
    ----------
    iterable : dict or array_like
        dict which defines the Fourier formula. Should be in form
        `{'PP':ndarray(...), 'PS':..., ...}`. Also accepts `array_like`
        of form `(('PP':ndarray(...)), ('PS', ...), ...)`.

    Notes
    -----
    In order to fit RA-SHG data in Fourier space, we need to know
    the different Fourier components of our fitting formula. An
    instance of the `fFormContainer` class provides a convenient
    means of holding and manipulating such a Fourier formula. It
    basically acts as a dictionary of `{pc:fformula}` pairs, where
    `pc` labels the polarization combination and `fformula` is an
    `ndarray` of length `2*M+1`. Here, the `nth` element of `fformula`
    contains a formula for the `(n-M)th` Fourier coefficient.
    """
    def __init__(self, iterable, M=16):
        self._M = M
        self._check_defining_iterable(iterable)
        self._fform_dict = self._copy_iterable(iterable)
        self._sympify()

    def _copy_iterable(self, iterable):
        _fform_dict = {
            k:np.array([s for s in v])
            for k,v in dict(iterable).items()
        }
        return _fform_dict

    def _check_defining_iterable(self, iterable):
        def raise_error():
            raise ValueError('Invalid Data input.')
        try:
            bad_key = False in [type(k) == str for k in dict(iterable).keys()]
            bad_val = False in [type(v) == np.ndarray and v.shape == (2*self._M+1,) and v.dtype == object and v.ndim == 1 for v in dict(iterable).values()]
        except:
            raise_error()
        if bad_key:
            raise_error()
        if bad_val:
            raise_error()

    def _sympify(self):
        for k in self.get_keys():
            for m in np.arange(-self._M, self._M+1):
                self._fform_dict[k][n2i(m, self._M)] = sp.sympify(self._fform_dict[k][n2i(m, self._M)])

    def get_free_symbols(self):
        """Return a sorted list of the free symbols in the Fourier formula."""
        free_symbols = []
        for k in self.get_keys():
            for m in np.arange(-self._M, self._M+1):
                for fs in self._fform_dict[k][n2i(m, self._M)].free_symbols:
                    free_symbols.append(fs)
        free_symbols = list(set(free_symbols))
        free_symbols.sort(key=str)
        return free_symbols

    def subs(self, subs_array):
        """Substitute for some variables in the Fourier formula.

        Parameters
        ----------
        subs_array : array_like of array_like of sympy.Expr
            An array of pairs of values. For example, to substitute the expression
            `2*xyz` for the variable `xxx`, use ``subs_array = ((xxx, 2*xyz))``.
        """
        subs_fform_dict = {}
        for k,v in self.get_items():
            subs_fform_dict[k] = np.zeros(shape=(2*self._M+1,), dtype=object)
            for m in np.arange(-self._M, self._M+1):
                subs_fform_dict[k][n2i(m, self._M)] = self._fform_dict[k][n2i(m, self._M)].subs(subs_array)
        self._fform_dict = subs_fform_dict

    def apply_phase_shift(self, phase_shift):
        """Phase shift the Fourier formula.
        
        Parameters
        ----------
        phase_shift : sympy.Expr 

        Notes
        -----
        This method works simply by multiplying the `mth` component of the
        Fourier formula by ``sympy.exp(1j * m * phase_shift)``.

        """
        new_fform_dict = {k:np.zeros(shape=(2*self._M+1,), dtype=object) for k in self.get_keys()}
        for k in self.get_keys():
            for m in np.arange(-self._M, self._M+1):
                new_fform_dict[k][n2i(m, self._M)] = self._fform_dict[k][n2i(m, self._M)] * (sp.cos(m * phase_shift) + 1j*sp.sin(m * phase_shift))
        self._fform_dict = new_fform_dict

    def get_M(self):
        """Get the number of Fourier frequencies.

        Each Fourier formula array has `2*M+1` elements."""
        return self._M

    def get_pc(self, pc):
        """Get the Fourier formula for single polarization combination.

        Parameters
        ----------
        pc : str
            Polarization combination of interest.

        Returns
        -------
        fformula : ndarray
            Fourier formula array for polarization combination `pc`. 
        """
        return np.copy(self._fform_dict[pc])
        
    def get_keys(self):
        """Get the polarization combinations for this Fourier formula.

        Equivalent to `dict.keys()`.
        
        Returns
        -------
        pcs : list of str
        """
        return list(self._fform_dict.keys())

    def get_values(self):
        """Get a list of the Fourier formulas (without polarization combination reference)
    
        Equivalent to `dict.values()`.

        Returns
        -------
        values : list of ndarray
            Data for each polarization combination
            in the form `array([xdata, ydata])`
        """
        return list(self._fform_dict.values())

    def get_items(self):
        """Get a list of `(pc, fformula)` pairs.

        Equivalent to `dict.items()`.

        Returns
        -------
        items : list of (str, ndarray) tuple
            List of data for each polarization combination, each
            in the form `(pc, array(shape=(2*M+1,)))`
        """
        return list(self._fform_dict.items())


class FormContainer:
    """Contains an SHG formula in `phi`-space.

    Parameters
    ----------
    iterable : dict or array_like
        dict which defines the formula. Should be in form
        `{'PP':sympy.Expr, 'PS':..., ...}`. Also accepts `array_like`
        of form `(('PP':sympy.Expr), ('PS', ...), ...)`.

    Notes
    -----
    While fitting RA-SHG in Fourier space is usually faster, for simple
    problems it is acceptable to fit in real space. An instance of the
    `FormContainer` class provides a convenient means of storing and
    manipulating such a formula. It basically acts as a dictionary of
    `{pc:formula}` pairs, where `pc` labels the polarization combination
    and `formula` is an instance of the sympy.Expr class.
    """

    def __init__(self, iterable):
        self._check_defining_iterable(iterable)
        self._form_dict = self._copy_iterable(iterable)
        self._sympify()

    def _copy_iterable(self, iterable):
        _form_dict = {
            k:v for k,v in dict(iterable).items()
        }
        return _form_dict

    def get_free_symbols(self):
        """Return a sorted list of the free symbols in the Fourier formula."""
        free_symbols = []
        for k in self.get_keys():
            for fs in self._form_dict[k].free_symbols:
                free_symbols.append(fs)
        free_symbols = list(set(free_symbols))
        free_symbols.sort(key=str)
        return free_symbols

    def subs(self, subs_array):
        """Substitute for some variables in the Fourier formula.

        Parameters
        ----------
        subs_array : array_like of array_like of sympy.Expr
            An array of pairs of values. For example, to substitute the expression
            `2*psi` for the variable `phi`, use ``subs_array = ((phi, 2*psi))``.
        """
        subs_form_dict = {}
        for k,v in self.get_items():
            subs_form_dict[k] = self._form_dict[k].subs(subs_array)
        self._form_dict = subs_form_dict
    
    def apply_phase_shift(self, phase_shift, var_to_shift):
        """Phase shift the formula.
        
        Parameters
        ----------
        phase_shift : sympy.Expr 
            Amount to shift by
        var_to_shift : sympy.Symbol
            Variable to shift

        Notes
        -----
        This method is equivalent to ``subs([var_to_shift, var_to_shift+phase_shift])``
        """
        self.subs([(var_to_shift, var_to_shift + phase_shift)])

    def _check_defining_iterable(self, iterable):
        def raise_error():
            raise ValueError('Invalid Data input.')
        try:
            bad_key = False in [type(k) == str for k in dict(iterable).keys()]
        except:
            raise_error()
        if bad_key:
            raise_error()

    def _sympify(self):
        for k in self.get_keys():
            self._form_dict[k] = sp.sympify(self._form_dict[k])

    def simplify(self):
        """Apply the `sympy.simplify` method to every polarization combination."""
        for k,v in self.get_items():
            self._form_dict[k] = sp.simplify(v)

    def expand(self):
        """Apply the `sympy.expand` method to every polarization combination."""
        for k,v in self.get_items():
            self._form_dict[k] = sp.expand(v)

    def get_keys(self):
        """Get the polarization combinations for this formula.

        Equivalent to `dict.keys()`.
        
        Returns
        -------
        pcs : list of str
        """
        return list(self._form_dict.keys())

    def get_values(self):
        """Get a list of the formulas (without polarization combination reference)
    
        Equivalent to `dict.values()`.

        Returns
        -------
        values : list of sympy.Expr
        """
        return list(self._form_dict.values())

    def get_items(self):
        """Get a list of `(pc, formula)` pairs.

        Equivalent to `dict.items()`.

        Returns
        -------
        items : list of (str, sympy.Expr) tuple
            List of data for each polarization combination, each
            in the form `(pc, formula)`
        """
        return list(self._form_dict.items())

    def get_pc(self, pc):
        """Get the formula for single polarization combination.

        Parameters
        ----------
        pc : str
            Polarization combination of interest.

        Returns
        -------
        formula : sympy.Expr
            Formula expression for polarization combination `pc`. 
        """
        return self._form_dict[pc]


def n2i(n, M=16):
    """Convert between Fourier index and array index

    Returns the index (0-`2*M+1`) in a Fourier array corresponding to the `nth` Fourier component
    """
    return n+M


def fform_to_fdat(fform, subs_dict):
    """Convert instance of `fFormContainer` to instance of `fDataContainer`

    Parameters
    ----------
    fform : fFormContainer
        Instance of :class:`~shgpy.core.data_handler.fFormContainer`
    subs_dict : dict
        Dict of `{sympy.Symbol : float}` pairs to perform the substitution.

    Returns
    -------
    fdat : fDataContainer
        Instance of `fDataContainer`

    Notes
    -----
    This function operates by taking an instance of the `fFormContainer`
    class and substituting for all the relevant `sympy.Symbol`s according
    to `subs_dict`.

    """
    subs_array = [(k,subs_dict[k]) for k in fform.get_free_symbols()]
    new_fform = fFormContainer(fform.get_items())
    new_fform.subs(subs_array)
    iterable = {k:v.astype(complex) for k,v in new_fform.get_items()}
    return fDataContainer(iterable, new_fform.get_M())
    

def fdat_to_dat(fdat, num_points): 
    """Convert instance of `fDataContainer` to instance of `DataContainer`

    Parameters
    ----------
    fdat : fDataContainer
        Instance of :class:`~shgpy.core.data_handler.fDataContainer`
    num_points : int
        Number of different points to put into the `DataContainer`.

    Returns
    -------
    dat : DataContainer
        Instance of :class:`~shgpy.core.data_handler.DataContainer`

    Notes
    -----
    This function operates by taking an instance of the `fDataContainer`
    class and performing an inverse Fourier transform.

    """
    iterable = {}
    M = fdat.get_M()
    xdata = np.linspace(0, 2*np.pi, num_points, endpoint=False)
    for k,v in fdat.get_items():
        ydata = np.zeros(num_points, dtype=complex)
        for m in np.arange(-M, M+1):
            ydata += v[n2i(m, M)] * (np.cos(m * xdata) + 1j*np.sin(m * xdata))
        iterable[k] = np.array([xdata, ydata]).real.astype(float)
    return DataContainer(iterable, 'radians', normal_to_oblique=False)


def fform_to_dat(fform, subs_dict, num_points):
    """Convert instance of `fFormContainer` to instance of `DataContainer`

    Parameters
    ----------
    fform : fFormContainer
        Instance of :class:`~shgpy.core.data_handler.fFormContainer`
    subs_dict : dict
        Dict of `{sympy.Symbol : float}` pairs to perform the substitution.
    num_points : int
        Number of different points to put into the `DataContainer`.

    Returns
    -------
    dat : DataContainer
        Instance of :class:`~shgpy.core.data_handler.DataContainer`

    Notes
    -----
    This function operates by taking an instance of the `fFormContainer`
    class and substituting for all the relevant `sympy.Symbol`s according
    to `subs_dict`. Then it creates a `DataContainer` instance with 
    `num_points` points.

    """
    return fdat_to_dat(fform_to_fdat(fform, subs_dict), num_points)


def fform_to_form(fform):
    """Convert instance of `fFormContainer` to instance of `FormContainer`

    Parameters
    ----------
    fform : fFormContainer
        Instance of :class:`~shgpy.core.data_handler.fFormContainer`

    Returns
    -------
    form : FormContainer
        Instance of :class:`~shgpy.core.data_handler.FormContainer`

    Notes
    -----
    This function operates by taking an instance of the `fFormContainer`
    class and performing a Fourier transform.

    """
    iterable = {}
    for k,v in fform.get_items():
        iterable[k] = _formula_from_fexpr(v, fform.get_M())
    return FormContainer(iterable)


def _formula_from_fexpr(t, M=16):
    expr = 0
    for m in np.arange(-M, M+1):
        expr += t[n2i(m)]*(sp.cos(m*S.phi)+1j*sp.sin(m*S.phi))
    return expr


def _fexpr_n(expr_arr, n, precision=7):
    Rshape = expr_arr.shape
    expr_arrf = expr_arr.flatten()
    h = np.zeros(len(expr_arrf), dtype=object)
    for i in range(len(expr_arrf)):
        f_re = sp.lambdify(S.phi, 1/2/sp.pi*expr_arrf[i]*sp.cos(-1*(n*S.phi)))
        f_im = sp.lambdify(S.phi, 1/2/sp.pi*expr_arrf[i]*sp.sin(-1*(n*S.phi)))
        t_re,_ = quad(f_re, 0, 2*np.pi)
        t_im,_ = quad(f_im, 0, 2*np.pi)
        h[i] = round(t_re, precision)+round(t_im, precision)*1j
    h = h.reshape(Rshape)
    return h


def form_to_dat(form, subs_dict, num_points):
    """Convert instance of `FormContainer` to instance of `DataContainer`

    Parameters
    ----------
    form : FormContainer
        Instance of :class:`~shgpy.core.data_handler.FormContainer`
    subs_dict : dict
        Dict of `{sympy.Symbol : float}` pairs to perform the substitution.
    num_points : int
        Number of different points to put into the `DataContainer`.

    Returns
    -------
    dat : DataContainer
        Instance of :class:`~shgpy.core.data_handler.DataContainer`

    Notes
    -----
    This function operates by taking an instance of the `FormContainer`
    class and substituting for all the relevant `sympy.Symbol`s according
    to `subs_dict`. Then it creates a `DataContainer` instance with 
    `num_points` points.

    """
    subs_array = [(k,subs_dict[k]) for k in subs_dict.keys()]
    new_form = FormContainer(form.get_items())
    new_form.subs(subs_array)
    if new_form.get_free_symbols() != [S.phi]:
        raise ValueError('Only one variable allowed in conversion from form to dat. Is subs_array correct?')
    iterable = {}
    for k,v in new_form.get_items():
        f = sp.lambdify(S.phi, v)
        xdata = np.linspace(0, 2*np.pi, num_points, endpoint=False)
        ydata = np.array([f(x) for x in xdata])
        iterable[k] = np.array([xdata, ydata], dtype=complex).real
    return DataContainer(iterable, 'radians', normal_to_oblique=False)


def _I_component(expr):
    return (expr-expr.subs(sp.I, 0)).subs(sp.I, 1)


def _no_I_component(expr):
    return expr.subs(sp.I, 0)


def form_to_fform(form, M=16):
    """Convert instance of `FormContainer` to instance of `fFormContainer`

    Parameters
    ----------
    form : FormContainer
        Instance of :class:`~shgpy.core.data_handler.FormContainer`
    M : int, optional
        Number of Fourier frequencies, defaults to `16`.

    Returns
    -------
    fform : fFormContainer
        Instance of :class:`~shgpy.core.data_handler.fFormContainer`

    Notes
    -----
    This function operates by taking an instance of the `FormContainer`
    class and performing a Fourier transform.

    """
    iterable = {}
    for k,v in form.get_items():
        iterable[k] = np.zeros((2*M+1,), dtype=object)
        _logger.info(f'Currently computing {k}.')
        for m in np.arange(-M, M+1):
            _logger.debug(f'Currently computing m={m}.')
            expr = sp.expand_trig(v*(sp.cos(-m*S.phi)+1j*sp.sin(-m*S.phi))).expand()
            expr_re = _no_I_component(expr)
            expr_im = _I_component(expr)
            iterable[k][n2i(m, M)] = 1/2/sp.pi*sp.integrate(expr_re, (S.phi, 0, 2*sp.pi)) + 1/2/sp.pi*sp.I*sp.integrate(expr_im, (S.phi, 0, 2*sp.pi))
    return fFormContainer(iterable, M=M)


def form_to_fdat(form, subs_dict, M=16):
    """Convert instance of `FormContainer` to instance of `fDataContainer`

    Parameters
    ----------
    fform : FormContainer
        Instance of :class:`~shgpy.core.data_handler.FormContainer`
    subs_dict : dict
        Dict of `{sympy.Symbol : float}` pairs to perform the substitution.

    Returns
    -------
    fdat : fDataContainer
        Instance of `fDataContainer`

    Notes
    -----
    This function operates by taking an instance of the `FormContainer`,
    doing a Fourier transform to a `fFormContainer` instance, and then
    substituting for all the relevant `sympy.Symbol`s according to
    `subs_dict`.

    """
    return fform_to_fdat(form_to_fform(form, M), subs_dict)


def merge_containers(containers, mapping):
    """Merge two contains with new keys according to `mapping`.

    Parameters
    ----------
    containers : list of Container_like
        List of Containers (DataContainer, fDataContainer,
        FormContainer, or fFormContainer). Must be the same type.
    mapping : function(key, index)
        Function which maps `key` of Container and `index`
        of a Container in `containers` to `str`.

    Returns
    -------
    container : Container_like
        Container object of type ``type(containers[0])`` with items
        ((k1, v1), (k2, v2), ...), where ki are the mapped keys of
        the containers and vi are the corresponding values.

    Examples
    --------
    >>> xdata = np.linspace(0, 2*np.pi, 1000)
    >>> ydata = np.sin(xdata)**2
    >>> iterable = {k:np.array([xdata, ydata]) for k in ['PP', 'PS']}
    >>> dat1 = DataContainer(iterable)
    >>> dat2 = DataContainer(iterable)
    >>> def mapping(key, index):
    >>>     return key+str(index)
    >>> dat3 = merge_containers([dat1, dat2], mapping)
    >>> dat3.get_keys():
    ['PP0', 'PS0', 'PP1', 'PS2']
    >>> dat3.get_values():
    [array([xdata, ydata]), array([xdata, ydata]), ...]

    """
    iterable = {}
    new_type = type(containers[0])
    if new_type == DataContainer:
        type_d = True
    else:
        type_d = False
    if new_type not in [DataContainer, fDataContainer, FormContainer, fFormContainer]:
        raise TypeError('Containers must be dype DataContainer, fDataContainer, FormContainer, or fFormContainer')
    for i,container in enumerate(containers):
        if type_d:
            items = container.get_items('radians')
        else:
            items = container.get_items()
        for k,v in items:
            new_k = mapping(k, i)
            new_v = np.copy(v)
            iterable[new_k] = new_v
            if type(container) != new_type:
                raise TypeError('Containers are of different types.')
    
    if type_d:
        return new_type(iterable, 'radians')
    else:
        return new_type(iterable)


def read_csv_file(filename, delimiter=','):
    """Read a csv file.

    Parameters
    ----------
    filename : str or file object
    delimiter : str
        Defaults to ``','``.

    Returns
    -------
    data : ndarray
        Data from csv file of the form `array([xdata, ydata])`.

    """
    xdata = []
    ydata = []
    with open(filename, 'r') as f:
        reader = csv.reader(f, delimiter=delimiter)
        for row in reader:
            xdata.append(float(row[0]))
            ydata.append(float(row[1]))
    return np.array([xdata, ydata])


def dat_subtract(dat1, dat2):
    """Subtract two `DataContainer` instances.

    Parameters
    ----------
    dat1 : DataContainer
        Instance of :class:`~shgpy.core.data_handler.DataContainer`
    dat2 : DataContainer
        Instance of :class:`~shgpy.core.data_handler.DataContainer`

    Returns
    -------
    dat_result : DataContainer
        Instance of :class:`~shgpy.core.data_handler.DataContainer` containing
        the contents of `dat1-dat2`.
    
    """
    if set(dat1.get_keys()) != set(dat2.get_keys()):
        raise ValueError('DataContainers have different keys.')
    if dat1.get_values('radians')[0].shape != dat2.get_values('radians')[0].shape:
        raise ValueError('DataContainers\' xydatas have different shapes.')
    new_dict = {}
    for k in dat1.get_keys():
        new_xdata, ydata1 = dat1.get_pc(k, 'radians')
        _, ydata2 = dat2.get_pc(k, 'radians')
        if False in np.isclose(new_xdata.flatten(), _.flatten()):
            raise ValueError('DataContainers have different xdatas.')
        new_ydata = ydata1 - ydata2
        new_dict[k] = np.array([new_xdata, new_ydata])
    return DataContainer(new_dict, 'radians', normal_to_oblique=False)
       

def load_data_and_dark_subtract(data_filenames_dict, data_angle_units, dark_filenames_dict, dark_angle_units, normal_to_oblique=False):
    """Load RA-SHG data from a dict of filenames and dark subtract.

    Parameters
    ----------
    data_filenames_dict : dict of (str : str or file object)
        dict of form (polarization combination : filename).
    data_angle_units : {'radians', 'degrees'}
        Units of `xdata`
    dark_filenames_dict : dict of (str : str or file object)
        dict of form (polarization combination : filename).
    dark_angle_units : {'radians', 'degrees'}
        Units of `xdata` for the dark.
    normal_to_oblique : bool, optional
        Defaults to False. See :class:`~shgpy.core.data_handler.DataContainer`.

    Returns
    -------
    dat : DataContainer
    
    """
    if set(data_filenames_dict.keys()) != set(dark_filenames_dict.keys()):
        raise ValueError('filename dicts have different keys.')
    dat1 = DataContainer({k:read_csv_file(v) for k,v in data_filenames_dict.items()}, data_angle_units, normal_to_oblique=normal_to_oblique)
    dat2 = DataContainer({k:read_csv_file(v) for k,v in dark_filenames_dict.items()}, dark_angle_units, normal_to_oblique=normal_to_oblique)
    return dat_subtract(dat1, dat2)
    

def load_data(data_filenames_dict, angle_units, normal_to_oblique=False):
    """Load RA-SHG data from a dict of filenames.

    Parameters
    ----------
    data_filenames_dict : dict of (str : str or file object)
        dict of form (polarization combination : filename).
    data_angle_units : {'radians', 'degrees'}
        Units of `xdata`
    normal_to_oblique : bool, optional
        Defaults to False. See :class:`~shgpy.core.data_handler.DataContainer`.

    Returns
    -------
    dat : DataContainer
    
    """
    return DataContainer({k:read_csv_file(v) for k,v in data_filenames_dict.items()}, angle_units, normal_to_oblique=normal_to_oblique)


def dat_to_fdat(dat, interp_kind='cubic', M=16):
    """Convert instance of `DataContainer` to instance of `fDataContainer`

    Parameters
    ----------
    dat : DataContainer
        Instance of :class:`~shgpy.core.data_handler.DataContainer`
    interp_kind : str or int, optional
        Specifies the kind of interpolation as a string (‘linear’,
        ‘nearest’, ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’, ‘previous’,
        ‘next’, where ‘zero’, ‘slinear’, ‘quadratic’ and ‘cubic’ refer
        to a spline interpolation of zeroth, first, second or third order;
        ‘previous’ and ‘next’ simply return the previous or next value of
        the point) or as an integer specifying the order of the spline
        interpolator to use. Default is ‘cubic’.
    M : int, optional
        Number of Fourier frequencies, defaults to `16`.

    Returns
    -------
    fdat : fDataContainer
        Instance of `fDataContainer`

    Notes
    -----
    This function works by taking a DataContainer instance and performing
    an interpolation so that the points are equally spaced along `phi`.
    Then it performs a Fourier transform and returns an fDataContainer
    instance. The interpolation is necessary because a large portion
    of the codebase requires equally spaced points in order for the Fourier
    transform to be well-defined.

    """
    ans = {pc:np.zeros(2*M+1, dtype=np.complex64) for pc in dat.get_keys()}
    for k in dat.get_keys():
        for m in np.arange(-M, M+1):
            xdata, ydata = dat.get_pc(k, 'radians')
            interp_func = interp1d(xdata, ydata, kind=interp_kind, fill_value='extrapolate')
            interp_xdata = np.linspace(0, 2*np.pi, len(ydata), endpoint=False)
            interp_ydata = interp_func(interp_xdata)
            dx = interp_xdata[1] - interp_xdata[0]
            ans[k][n2i(m, M)] = sum([1/2/np.pi*dx*interp_ydata[i]*np.exp(-1j*m*interp_xdata[i]) for i in range(len(interp_xdata))])
    return fDataContainer(ans.items(), M=M)


def load_data_and_fourier_transform(data_filenames_dict, data_angle_units, dark_filenames_dict=None, dark_angle_units=None, interp_kind='cubic', M=16, min_subtract=False, scale=1, normal_to_oblique=False):
    """Load RA-SHG data from a dict of filenames and Fourier transform

    Parameters
    ----------
    data_filenames_dict : dict of (str : str or file object)
        dict of form (polarization combination : filename).
    data_angle_units : {'radians', 'degrees'}
        Units of `xdata`
    dark_filenames_dict : dict of (str : str or file object)
        dict of form (polarization combination : filename).
    dark_angle_units : {'radians', 'degrees'}
        Units of `xdata` for the dark.
    interp_kind : str or int, optional
        Specifies the kind of interpolation as a string (‘linear’,
        ‘nearest’, ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’, ‘previous’,
        ‘next’, where ‘zero’, ‘slinear’, ‘quadratic’ and ‘cubic’ refer
        to a spline interpolation of zeroth, first, second or third order;
        ‘previous’ and ‘next’ simply return the previous or next value of
        the point) or as an integer specifying the order of the spline
        interpolator to use. Default is ‘cubic’
        Dict of `{sympy.Symbol : float}` pairs to perform the substitution.
    M : int, optional
        Number of Fourier frequencies, defaults to `16`.
    min_subtract : bool
        Whether to subtract the minimum before Fourier transforming.
        Default is False.
    normal_to_oblique : bool, optional
        Defaults to False. See :class:`~shgpy.core.data_handler.DataContainer`.

    Returns
    -------
    dat : DataContainer
    fdat : fDataContainer
    
    """
    if dark_filenames_dict is not None:
        dat = load_data_and_dark_subtract(data_filenames_dict, data_angle_units, dark_filenames_dict, dark_angle_units, normal_to_oblique=normal_to_oblique)
    else:
        dat = load_data(data_filenames_dict, data_angle_units, normal_to_oblique=normal_to_oblique)
    if min_subtract:
        dat.subtract_min()
    fdat = dat_to_fdat(dat, interp_kind=interp_kind, M=M)
    if scale != 1:
        dat.scale_data(scale)
        fdat.scale_data(scale)
    return dat, fdat


def load_fform(fform_filename):
    """Creat instance of fFormContainer from a fform_filename.

    ``fform_filename`` s are generated using the utilities provided
    in the :class:`~shgpy.fformgen` package. Those utilities output
    a pickled fFormContainer-like object which can be loaded into
    a true fFormContainer instance using this function.

    Parameters
    ----------
    fform_filename : str or file object

    """
    with open(fform_filename, 'rb') as f:
        fform_dict = pickle.load(f)
    return fFormContainer(fform_dict)



import unittest

import logging

import shgpy
from shgpy.fformfit import (
    least_squares_fit,
    least_squares_fit_with_bounds,
    basinhopping_fit,
    basinhopping_fit_with_bounds,
    basinhopping_fit_jac,
    basinhopping_fit_jac_with_bounds,
)
from shgpy.plotter import easy_plot, easy_polar_plot
import numpy as np
import shgpy.shg_symbols as S

logger = logging.getLogger(__name__)

MANUAL = False


class TestFit(unittest.TestCase):

    fform_filename = 'tests/fform/T_d-S_2-S_2(110)-particularized.p'
    fform = shgpy.load_fform(fform_filename)
    data_filenames_dict = {
        'PP':'tests/Data/dataPP.csv',
        'PS':'tests/Data/dataPS.csv',
        'SP':'tests/Data/dataSP.csv',
        'SS':'tests/Data/dataSS.csv',
    }
    dat, fdat = shgpy.load_data_and_fourier_transform(data_filenames_dict, 'degrees')
    fform.apply_phase_shift(S.psi)
    guess_dict = {k:1 for k in fform.get_free_symbols()}
    bounds_dict = {k:((-2, 2) if k != S.psi else (-np.pi, np.pi)) for k in fform.get_free_symbols()}

    def test_least_squares(self):
        ret1 = least_squares_fit(self.fform, self.fdat, self.guess_dict)
        ret2 = least_squares_fit_with_bounds(self.fform, self.fdat, self.guess_dict, self.bounds_dict)
        for ret in [ret1, ret2]:
            self.assertAlmostEqual(abs(ret.xdict[S.psi]), 1.59, delta=0.1)
            self.assertAlmostEqual(abs(ret.xdict[S.zyx]), 1.23, delta=0.1)

        fit_dat = shgpy.fform_to_dat(self.fform, ret1.xdict, 1000)
        if MANUAL:
            easy_plot([self.dat, fit_dat], [{'linestyle':' ', 'markerfacecolor':'none', 'color':'blue', 'marker':'o'}, {'linestyle':'-', 'color':'blue'}], ['PP', 'PS', 'SP', 'SS'], show_plot=True, filename=None, show_legend=False)
            easy_polar_plot([self.dat, fit_dat], [{'linestyle':' ', 'markerfacecolor':'none', 'color':'blue', 'marker':'o'}, {'linestyle':'-', 'color':'blue'}], ['PP', 'PS', 'SP', 'SS'], show_plot=True, filename=None, show_legend=False)

    def test_basinhopping(self):
        niter = 100
        ret1 = basinhopping_fit(self.fform, self.fdat, self.guess_dict, niter)
        ret2 = basinhopping_fit_jac(self.fform, self.fdat, self.guess_dict, niter)
        ret3 = basinhopping_fit_with_bounds(self.fform, self.fdat, self.guess_dict, self.bounds_dict, niter)
        ret4 = basinhopping_fit_jac_with_bounds(self.fform, self.fdat, self.guess_dict, self.bounds_dict, niter)
        for ret in [ret1, ret2, ret3, ret4]:
            self.assertAlmostEqual(abs(ret.xdict[S.psi]), 1.59, delta=0.1)
            self.assertAlmostEqual(abs(ret.xdict[S.zyx]), 1.23, delta=0.1)



import unittest

import shgpy
import numpy as np
import logging
import random
import shgpy.shg_symbols as S
from shgpy.plotter import easy_plot

_logger = logging.getLogger(__name__)

MANUAL = False


class TestData(unittest.TestCase):

    def test_creation(self):
        items = (('PS', np.arange(0, 4).reshape((2,2))),('SS', np.arange(0, 4).reshape((2,2))),)
        hi = shgpy.DataContainer(items, 'radians')
        self.assertIsInstance(hi, shgpy.DataContainer)

    def test_dict_funcs(self):
        items = (('PP', np.arange(0, 4).reshape((2,2))),('PS', np.arange(0, 4).reshape((2,2))),('SP', np.arange(0, 4).reshape((2,2))),('SS', np.arange(0, 4).reshape((2,2))),)
        hi = shgpy.DataContainer(items, 'radians')
        for i,j in zip(hi.get_items('radians'), list(items)):
            self.assertEqual(i[0], j[0])
            self.assertTrue(np.array_equal(i[1], j[1]))

    def test_manipulation(self):
        items = (('PP', np.arange(0, 4).reshape((2,2))),('PS', np.arange(0, 4).reshape((2,2))),)
        hi = shgpy.DataContainer(items, 'radians')
        hi.scale_data(100)
        self.assertEqual(hi.get_scale(), 100)
        hi.phase_shift_data(10, angle_units='degreeS')
        self.assertEqual(hi.get_phase_shift('radians'), np.deg2rad(10))

    def test_maxval(self):
        items = (('PP', np.zeros(shape=(3,3))), ('PS', np.ones(shape=(4,4))))
        hi = shgpy.DataContainer(items, 'degrees')
        pc, maxval = hi.get_maxval()
        self.assertEqual((pc, maxval), ('PS', 1))


class TestfData(unittest.TestCase):

    def test_creation(self):
        M = 16
        items = (('PP', np.zeros(2*M+1, dtype=complex)), ('PS', np.ones(2*M+1, dtype=complex)*4))
        hi = shgpy.fDataContainer(items, M=M)
        self.assertIsInstance(hi, shgpy.fDataContainer)

    def test_dict_funcs(self):
        M = 16
        items = (('PP', np.zeros(2*M+1, dtype=complex)), ('PS', np.ones(2*M+1, dtype=complex)*4))
        hi = shgpy.fDataContainer(items, M=M)
        self.assertEqual(hi.get_keys(), list(dict(items).keys()))
        for i,j in zip(hi.get_items(), list(dict(items).items())):
            for k, l in zip(i, j):
                for a, b in zip(k, l):
                    self.assertEqual(a, b)

    def test_manipulation(self):
        M = 16
        items = (('PP', np.zeros(2*M+1, dtype=complex)), ('PS', np.ones(2*M+1, dtype=complex)*4))
        hi = shgpy.fDataContainer(items, M=M)
        hi.scale_fdata(14)
        self.assertEqual(hi.get_scale(), 14)
        hi.normalize_fdata(1)
        self.assertEqual(hi.get_scale(), 0.25)
        self.assertEqual(hi.get_maxval(), ('PS', 1.0))
        hi.phase_shift_fdata(14, 'degrees')
        self.assertEqual(hi.get_phase_shift('radians'), np.deg2rad(14))


class TestLoadData(unittest.TestCase):

    filenames_dict = {
        'PP':'tests/Data/dataPP.csv',
        'PS':'tests/Data/dataPS.csv',
        'SP':'tests/Data/dataSP.csv',
        'SS':'tests/Data/dataSS.csv',
    }
    
    def test_loaders(self):
        dat = shgpy.load_data(self.filenames_dict, 'degrees')
        self.assertIsInstance(dat, shgpy.DataContainer)
        dat = shgpy.load_data_and_dark_subtract(self.filenames_dict, 'degrees', self.filenames_dict, 'degrees')
        self.assertIsInstance(dat, shgpy.DataContainer)
        self.assertEqual(dat.get_maxval()[1], 0)
        dat, fdat = shgpy.load_data_and_fourier_transform(self.filenames_dict, 'degrees')
        self.assertNotEqual(fdat.get_maxval()[1], 0)
        dat, fdat = shgpy.load_data_and_fourier_transform(self.filenames_dict, 'degrees', self.filenames_dict, 'degrees')
        self.assertEqual(fdat.get_maxval()[1], 0)
    

class TestFormAndfForm(unittest.TestCase):

    fform_filename = 'tests/fform/T_d-S_2-S_2(110).p'
    fform = shgpy.load_fform(fform_filename)
    fform.apply_phase_shift(S.psi)

    def test_fform_to_form(self):
        _logger.debug(shgpy.fform_to_form(self.fform).get_items())

    def test_form_to_dat(self):
        form = shgpy.fform_to_form(self.fform)
        dat = shgpy.form_to_dat(form, {k:random.uniform(-1, 1) for k in form.get_free_symbols() if k != S.phi}, 1000)
        if MANUAL:
            easy_plot([dat], [{'linestyle':'-', 'color':'blue'}], dat.get_keys())


class TestConversions(unittest.TestCase):

    def random_subs_dict(self, keys):
        ans = {}
        for k in keys:
            ans[k] = random.uniform(-1, 1)
        return ans

    def test_convert(self):

        fform_filename = 'tests/fform/T_d-S_2-S_2(110).p'
        fform = shgpy.load_fform(fform_filename)
        data_filenames_dict = {
            'PP':'tests/Data/dataPP.csv',
            'PS':'tests/Data/dataPS.csv',
            'SP':'tests/Data/dataSP.csv',
            'SS':'tests/Data/dataSS.csv',
        }
        dat, fdat = shgpy.load_data_and_fourier_transform(data_filenames_dict, 'degrees')
        form = shgpy.fform_to_form(fform)

        t1 = shgpy.dat_to_fdat(dat)
        self.assertIsInstance(t1, type(fdat))
        t2 = shgpy.fdat_to_dat(fdat, 1000)
        self.assertIsInstance(t2, type(dat))
        t3 = shgpy.form_to_dat(form, self.random_subs_dict(fform.get_free_symbols()), 1000)
        self.assertIsInstance(t3, type(dat))
        t4 = shgpy.form_to_fform(form)
        self.assertIsInstance(t4, type(fform))
        t5 = shgpy.form_to_fdat(form, self.random_subs_dict(fform.get_free_symbols()))
        self.assertIsInstance(t5, type(fdat))
        t6 = shgpy.fform_to_fdat(fform, self.random_subs_dict(fform.get_free_symbols()))
        self.assertIsInstance(t6, type(fdat))
        t7 = shgpy.fform_to_dat(fform, self.random_subs_dict(fform.get_free_symbols()), 1000)
        self.assertIsInstance(t7, type(dat))
        t8 = shgpy.fform_to_form(fform)
        self.assertIsInstance(t8, type(form))


class TestNormalIncidence(unittest.TestCase):

    def test_normal_loading(self):

        data_filenames_dict = {
            'PP':'tests/Data/dataPP.csv',
            'SP':'tests/Data/dataPS.csv',
        }
        dat = shgpy.load_data(data_filenames_dict, 'degrees')
        if MANUAL:
            easy_plot([dat], [{'linestyle':'-', 'marker':'o', 'markerfacecolor':'none', 'color':'blue'}], dat.get_keys())
        dat, fdat = shgpy.load_data_and_fourier_transform(data_filenames_dict, 'degrees')


class TestPathologicalData(unittest.TestCase):

    filenames_dict = {
        'PP':'tests/Data/dataPP_pathological.csv',
        'PS':'tests/Data/dataPS_pathological.csv',
        'SP':'tests/Data/dataSP_pathological.csv',
        'SS':'tests/Data/dataSS_pathological.csv',
    }
    
    def test_loaders(self):

        dat = shgpy.load_data(self.filenames_dict, 'degrees')
        self.assertIsInstance(dat, shgpy.DataContainer)
        dat = shgpy.load_data_and_dark_subtract(self.filenames_dict, 'degrees', self.filenames_dict, 'degrees')
        self.assertIsInstance(dat, shgpy.DataContainer)
        self.assertEqual(dat.get_maxval()[1], 0)
        dat, fdat = shgpy.load_data_and_fourier_transform(self.filenames_dict, 'degrees')
        self.assertNotEqual(fdat.get_maxval()[1], 0)
        dat, fdat = shgpy.load_data_and_fourier_transform(self.filenames_dict, 'degrees', self.filenames_dict, 'degrees')
        self.assertEqual(fdat.get_maxval()[1], 0)
        












import unittest

import shgpy
import numpy as np


class TestMergeContainers(unittest.TestCase):

    data_filenames_dict = {
        'PP':'tests/Data/dataPP.csv',
        'PS':'tests/Data/dataPS.csv',
        'SP':'tests/Data/dataSP.csv',
        'SS':'tests/Data/dataSS.csv',
    }
    fform_filename = 'tests/fform/T_d-S_2-S_2(110).p'

    def test_dat(self):
        dat1 = shgpy.load_data(self.data_filenames_dict, 'degrees')
        dat2 = shgpy.load_data(self.data_filenames_dict, 'degrees')

        def mapping(key, index):
            return str(index) + key

        dat3 = shgpy.merge_containers([dat1, dat2], mapping)
        iterable = {}
        for k,v in dat1.get_items('radians'):
            iterable['0'+k] = np.copy(v)
        for k,v in dat2.get_items('radians'):
            iterable['1'+k] = np.copy(v)
        dat3_compare = shgpy.DataContainer(iterable, 'radians')

        for k in dat3.get_keys():
            a = dat3.get_pc(k, 'radians')
            b = dat3_compare.get_pc(k, 'radians')
            for i in range(2):
                self.assertEqual(str(a[i]), str(b[i]))

    def test_fform(self):
        fform1 = shgpy.load_fform(self.fform_filename)
        fform2 = shgpy.load_fform(self.fform_filename)

        def mapping(key, index):
            return str(index) + key

        fform3 = shgpy.merge_containers([fform1, fform2], mapping)
        iterable = {}
        for k,v in fform1.get_items():
            iterable['0'+k] = np.copy(v)
        for k,v in fform2.get_items():
            iterable['1'+k] = np.copy(v)
        fform3_compare = shgpy.fFormContainer(iterable)

        for k in fform3.get_keys():
            a = fform3.get_pc(k)
            b = fform3_compare.get_pc(k)
            self.assertEqual(str(a), str(b))




import unittest

import numpy as np
import sympy as sp
import logging

import shgpy

_logger = logging.getLogger(__name__)


class TestDeepCopyfForm(unittest.TestCase):

    def test_deepcopy_fform(self):
        M = 16
        x = sp.symbols('x', real=True)
        iterable = {}
        for k in ['PP', 'PS', 'SP', 'SS']:
            iterable[k] = np.zeros(2*M+1, dtype=sp.Expr)
            for i in np.arange(-16, 17):
                iterable[k][i] = x**2
        fform = shgpy.fFormContainer(iterable)
        self.assertNotEqual(fform.get_pc('SS')[16].subs(sp.symbols('x'), 1), 1)
        self.assertEqual(fform.get_pc('SS')[16].subs(sp.symbols('x', real=True), 1), 1)
        self.assertEqual(iterable['SS'][16], x**2)


class TestDeepCopyForm(unittest.TestCase):

    def test_deepcopy_form(self):
        x = sp.symbols('x', real=True)
        iterable = {}
        for k in ['PP', 'PS', 'SP', 'SS']:
            iterable[k] = x**2
        form = shgpy.FormContainer(iterable)
        self.assertNotEqual(form.get_pc('SS').subs(sp.symbols('x'), 1), 1)
        self.assertEqual(form.get_pc('SS').subs(sp.symbols('x', real=True), 1), 1)
        self.assertEqual(iterable['SS'], x**2)



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

with open('.version', 'r') as fh:
    version = fh.read().splitlines()[0]

setuptools.setup(
    name='shgpy',
    version=version,
    author='Bryan Fichera',
    author_email='bfichera@mit.edu',
    description='A collection of utilities for analyzing SHG data.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://bfichera.github.io/shgpy/',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



import numpy as np
import sympy as sp
from sympy.solvers import solve
import itertools
from warnings import warn


def particularize(tensor, exclude=[]):
    """Particularize `tensor` (e.g. enforce ``chi_ijk = chi_ikj``).

    Because (e.g.) ``P_i = chi_ijk E_j E_k``, ``chi_ijk`` needs to be symmetric
    in its last two indices. This symmetry is not documented in SHG tables,
    so it needs to be implemented manually. Functions using the ``solve``
    function of sympy.solvers.

    Parameters
    ----------
    tensor : ndarray
        Tensor to be particularized. 

    exclude : list of sympy.Symbol objects
        Variables to exlude from the solver. Defaults to ``[]``.

    Returns
    -------
    particularized_tensor : ndarray

    """

    def _swap_last_two_indices(tensor):
        rank = len(tensor.shape)
        dim = len(tensor[0])
        ans = np.zeros(shape=tensor.shape, dtype=tensor.dtype)
        for i in itertools.product(*[range(dim) for s in range(rank-2)]):
            for j in itertools.combinations_with_replacement(range(3), 2):
                ans[i][j] = tensor[i][(j[1], j[0])]
                ans[i][(j[1], j[0])] = tensor[i][j]
        return ans

    def _get_sols_from_particularization(tensor, exclude=[]):
        tensor2 = _swap_last_two_indices(tensor)
        sols = solve((tensor-tensor2).flatten(), dict=True, exclude=exclude)
        return sols

    def _apply_sol_to_tensor(tensor, sol):
        return np.array([fa.subs(sol) for fa in tensor.flatten()]).reshape(tensor.shape)

    if np.array_equal(_swap_last_two_indices(tensor), tensor):
        return tensor

    return _apply_sol_to_tensor(tensor, _get_sols_from_particularization(tensor, exclude=exclude)[0])


def _make_parameters_in_expr_complex(expr, prefix=('real_', 'imag_'), suffix=('', '')):
    free_symbols = expr.free_symbols
    for fs in free_symbols:
        if fs.is_real is None:
            expr = expr.subs(str(fs), sp.symbols(prefix[0]+str(fs)+suffix[0], real=True)+sp.I*sp.symbols(prefix[1]+str(fs)+suffix[1], real=True))
    return expr


def make_tensor_complex(tensor, prefix=('real_', 'imag_'), suffix=('', '')):
    """Substitute e.g. ``x`` in sympy expression with ``real_x+1j*imag_x``.

    In sympy, variables initalized by ``x = sympy.symbols('x')`` are by
    default assumed to be complex. In order to make this more explicit
    (e.g. for
    :func:`~shgpy.fformgen.generate_contracted_fourier_transforms_complex`),
    we replace ``x`` by ``real_x + 1j*imag_x``.


    Parameters
    ----------
    tensor : ndarray
    prefix : tuple of str, optional
        The prefixes of the newly-created real and imaginary variables.
        Defaults to ``('real_', 'imag_')``.
    suffix : tuple of str, optional
        The prefixes of the newly-created real and imaginary variables.
        Defaults to ``('', '')``.

    Returns
    -------
    complex_tensor : ndarray
    """
    shape = tensor.shape
    tensor = tensor.flatten()
    for i in range(len(tensor)):
        tensor[i] = _make_parameters_in_expr_complex(sp.sympify(tensor[i]), prefix=prefix, suffix=suffix)
    return np.reshape(tensor, shape)


def rotation_matrix3(n, t):
    """Returns the 3x3 matrix which rotates by `t` about `n`

    Parameters
    ----------
    n : ndarray
        Axis of rotation (ndim = 3).
    t : int, float, ...
        Angle (in radians) to rotate by.

    Returns
    -------
    rotation_matrix3 : ndarray of float
    """
    n_mag = _norm(n)
    for i in range(3):
        n[i] /= n_mag
    c = np.cos(t)
    s = np.sin(t)
    nx,ny,nz = n
    row1 = [c+nx**2*(1-c), nx*ny*(1-c)-nz*s, nx*nz*(1-c) + ny*s]
    row2 = [ny*nx*(1-c)+nz*s, c+ny**2*(1-c), ny*nz*(1-c)-nx*s]
    row3 = [nz*nx*(1-c) - ny*s, nz*ny*(1-c) + nx*s, c+nz**2*(1-c)]
    return np.array([row1, row2, row3])


def rotation_matrix3symb(n,t, ndigits=16):
    """Duplicate of :func:`~shgpy.core.utilities.rotation_matrix3`, but accepts sympy.Symbol angle argument.

    Parameters
    ----------
    n : ndarray
        Axis of rotation (ndim = 3).
    t : sympy.Symbol, ...
        Angle to rotate by.

    Returns
    -------
    rotation_matrix3symb : ndarray of sympy.Expr
    """
    n_mag = _normsymb(n)
    for i in range(3):
        n[i] /= n_mag
    c = sp.cos(t)
    s = sp.sin(t)
    nx,ny,nz = n
    row1 = [c+nx**2*(1-c), nx*ny*(1-c)-nz*s, nx*nz*(1-c) + ny*s]
    row2 = [ny*nx*(1-c)+nz*s, c+ny**2*(1-c), ny*nz*(1-c)-nx*s]
    row3 = [nz*nx*(1-c) - ny*s, nz*ny*(1-c) + nx*s, c+nz**2*(1-c)]
    ans = np.array([row1, row2, row3], dtype=object)
    return ans


def _levi_civita(i, j, k, first_index=0):
    i -= first_index
    j -= first_index
    k -= first_index
    ans = 0
    pos_args = [[0, 1, 2], [1, 2, 0], [2, 0, 1]]
    neg_args = [[1, 0, 2], [0, 2, 1], [2, 1, 0]]
    if [i, j, k] in pos_args:
        ans = 1
    if [i, j, k] in neg_args:
        ans = -1
    return ans


def _cross_product(v1, v2):
    ans = [0, 0, 0]
    for i in range(3):
        for j in range(3):
            for k in range(3):
                ans[i] += _levi_civita(i, j, k)*v1[j]*v2[k]
    return np.array(ans)


def _norm(v):
    return np.sqrt(v[0]**2 + v[1]**2 + v[2]**2)


def _normsymb(v):
    return sp.sqrt(v[0]**2 + v[1]**2 + v[2]**2)


def rotation_matrix_from_two_vectors(v_initial, v_final, accuracy=.01, ndigits=16):
    """Return rotation matrix which takes `v_initial` -> `v_final`.

    Parameters
    ----------
    v_initial : ndarray
    v_final : ndarray
    accuracy : float, optional
        Precision to which `v_final` is achieved. Defaults to `0.1`.
    ndigits : int, optional
        Number of digits to round off to. Defaults to `16`.

    Returns
    -------
    rotation_matrix3 : ndarray of float
    """
    v_initial = v_initial.astype(float)
    v_final = v_final.astype(float)
    iden = np.identity(3)
    mag_v_initial = _norm(v_initial)
    mag_v_final = _norm(v_final)
    for i in range(3):
        v_initial[i] /= mag_v_initial
        v_final[i] /= mag_v_final
    cos_theta = sum([v_initial[i]*v_final[i] for i in range(3)])
    mag_theta = np.arccos(cos_theta)
    axis = _cross_product(v_initial, v_final)
    done = False
    if np.array_equal(axis, np.zeros(3, dtype=axis.dtype)):
        if np.array_equal(-v_initial, v_final):
            ans = -iden
            done = True
        elif np.array_equal(v_initial, v_final):
            ans = iden
            done = True
    if not done:
        ans = rotation_matrix3(axis, mag_theta)
        mag_axis = _norm(axis)
        for i in range(3):
            axis[i] /= mag_axis
        check = [np.dot(ans, v_initial)[i] - v_final[i] < accuracy for i in range(3)]
        if False in check:
            ans = rotation_matrix3(axis, -mag_theta)
            check = [np.dot(ans, v_initial)[i] - v_final[i] < accuracy for i in range(3)]
            if False in check:
                warn('Rotation matrix from two vectors failed precision check.')

    for i in range(len(ans)):
        for j in range(len(ans[0])):
            ans[i][j] = round(ans[i][j], ndigits)
    return np.array(ans)


def union(*arrays):
    """Union of multiple arrays."""
    ans = []
    for y in arrays:
        for x in y:
            if x not in ans:
                ans.append(x)
    return ans


def tensor_product(*tensors):
    """Tensor product of multiple tensors."""
    if len(tensors) == 2:
        return np.tensordot(tensors[0], tensors[1], axes=0)
    else:
        return np.tensordot(tensors[0], tensor_product(*tensors[1:]), axes=0)


def tensor_contract(tensor, index_pairs):
    """Contract a tensor against to pairs of indices.

    Parameters
    ----------
    tensor : ndarray
    index_pairs : array_like of array_like of int
        List of pairs of indices, e.g. ``[[1, 3], [2, 4]]``.
    """
    ans = tensor
    index_idx = list(range(len(np.shape(tensor))))
    index_pairs = index_pairs[:]
    while len(index_pairs) > 0:
        old_index_pair = index_pairs.pop(0)
        new_index_pair = [index_idx.index(old_index_pair[i]) for i in range(len(old_index_pair))]
        ans = np.trace(ans, axis1=new_index_pair[0], axis2=new_index_pair[1])
        for i in range(len(old_index_pair)):
            index_idx.remove(old_index_pair[i])
    return ans
    

def transform(tensor, operation):
    """Transform a tensor by a given operation.

    Parameters
    ----------
    tensor : ndarray
    operation : ndarray
        Operation to transform `tensor` by. Should be rank 2.

    Returns
    -------
    transformed_tensor : ndarray
    """
    rank = len(tensor.shape)
    args = [operation]*rank
    args.append(tensor)
    return tensor_contract(tensor_product(*args), [[2*i+1, 2*rank+i] for i in range(rank)])



"""Module for (non-Fourier) formula derivation

This module provides a number of routines for calculating the SHG 
response of a given material (encoded by its susceptbility tensor,
see :class:`~shgpy.tensor_definitions` ). These routines are much
more straightforward than those found in :class:`~shgpy.fformgen`,
but are only suited for the most simple problems because the
conversion from formulas to Fourier formulas (which is all but
necessary for efficient fitting functionality) is typically quite
slow.

"""
import numpy as np
import sympy as sp
from .core import (
    FormContainer,
    tensor_contract,
    tensor_product,
)
from . import shg_symbols as S


# TODO
# Make all the functions here use gen_P_...


def make_form_from_P_and_Q(Pp, Ps, Qp, Qs, mode):
    """Given a dipole and quadrupole moment, return the SHG signal.

    Parameters
    ----------
    Pp : sympy.Expr
        2\\omega dipole component if the input is P polarized
    Ps : sympy.Expr
        2\\omega dipole component if the input is S polarized
    Qp : sympy.Expr
        2\\omega quadrupole component if the input is P polarized
    Qs : sympy.Expr
        2\\omega quadrupole component if the input is S polarized
    mode : {'real', 'complex'}
        Whether to treat the sympy.Symbols as real or complex.

    Returns
    -------
    form : FormContainer

    """
    if mode == 'complex':

        Sp = Pp + sp.I*Qp
        Ss = Ps + sp.I*Qs

        PP = Sp[0]*sp.conjugate(Sp[0])+Sp[2]*sp.conjugate(Sp[2])
        PS = Sp[1]*sp.conjugate(Sp[1])
        SP = Ss[0]*sp.conjugate(Ss[0])+Ss[2]*sp.conjugate(Ss[2])
        SS = Ss[1]*sp.conjugate(Ss[1])

    elif mode == 'real':

        PP = Pp[0]**2+Pp[2]**2+Qp[0]**2+Qp[2]**2
        PS = Pp[1]**2+Qp[1]**2
        SP = Ps[0]**2+Ps[2]**2+Qs[0]**2+Qs[2]**2
        SS = Ps[1]**2+Qs[1]**2

    else:

        raise ValueError('mode must be either \'real\' or \'complex\'')

    return FormContainer({'PP':PP, 'PS':PS, 'SP':SP, 'SS':SS})


def gen_P_just_dipole(t1, theta):
    """Generate P formula assuming dipole SHG.

    Parameters
    ----------
    t1 : ndarray of sympy.Expr
        SHG susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    theta : float or sympy.Symbol
        Angle of incidence

    Returns
    -------
    Pp : array_like of sympy.Expr
        The 2\\omega polarization assuming P-polarized input
    Ps : array_like of sympy.Expr
        The 2\\omega polarization assuming S-polarized input

    """
    c = sp.cos(theta)
    s = sp.sin(theta)
    kout = np.array([s, 0, -c], dtype=object)
    Fp = np.array([-c, 0, s], dtype=object)
    Fs = np.array([0, 1, 0], dtype=object)
    R = np.array([[sp.cos(S.phi), -sp.sin(S.phi), 0], [sp.sin(S.phi), sp.cos(S.phi), 0], [0, 0, 1]])
    rotated_tensor = tensor_contract(tensor_product(R, R, R, t1), [[1, 6], [3, 7], [5, 8]])
    Ps = tensor_contract(tensor_product(rotated_tensor, Fs, Fs), [[1, 3], [2, 4]])
    Pp = tensor_contract(tensor_product(rotated_tensor, Fp, Fp), [[1, 3], [2, 4]])
    Ps -= np.dot(kout, Ps)*kout
    Pp -= np.dot(kout, Pp)*kout

    return Pp, Ps, 0, 0


def gen_P_dipole_quadrupole(t1, t2, theta):
    """Generate P formula assuming dipole+quadrupole SHG.

    Parameters
    ----------
    t1 : ndarray of sympy.Expr
        SHG dipole susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    t2 : ndarray of sympy.Expr
        SHG quadrupole susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    theta : float or sympy.Symbol
        Angle of incidence

    Returns
    -------
    Pp_eff : array_like of sympy.Expr
        The 2\\omega effective polarization assuming P-polarized input
    Ps_eff : array_like of sympy.Expr
        The 2\\omega effective polarization assuming S-polarized input
    
    """
    c = sp.cos(theta)
    s = sp.sin(theta)
    kin = np.array([s, 0, c], dtype=object)
    kout = np.array([s, 0, -c], dtype=object)
    Fp = np.array([-c, 0, s], dtype=object)
    Fs = np.array([0, 1, 0], dtype=object)
    R = np.array([[sp.cos(S.phi), -sp.sin(S.phi), 0], [sp.sin(S.phi), sp.cos(S.phi), 0], [0, 0, 1]])
    rotated_tensor = tensor_contract(tensor_product(R, R, R, t1), [[1, 6], [3, 7], [5, 8]])
    rotated_qtensor = tensor_contract(tensor_product(R, R, R, R, t2), [[1, 8], [3, 9], [5, 10], [7, 11]])
    Ps = tensor_contract(tensor_product(rotated_tensor, Fs, Fs), [[1, 3], [2, 4]])
    Pp = tensor_contract(tensor_product(rotated_tensor, Fp, Fp), [[1, 3], [2, 4]])
    Qs = tensor_contract(tensor_product(rotated_qtensor, kin, Fs, Fs), [[1, 4], [2, 5], [3, 6]])
    Qp = tensor_contract(tensor_product(rotated_qtensor, kin, Fp, Fp), [[1, 4], [2, 5], [3, 6]])
    Ps -= np.dot(kout, Ps)*kout
    Pp -= np.dot(kout, Pp)*kout
    Qs -= np.dot(kout, Qs)*kout
    Qp -= np.dot(kout, Qp)*kout
    return Pp, Ps, Qp, Qs


def formgen_just_dipole_complex(t1, theta):
    """Generate formula assuming dipole SHG with complex coefficients.

    Parameters
    ----------
    t1 : ndarray of sympy.Expr
        SHG susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    theta : float or sympy.Symbol
        Angle of incidence

    Returns
    -------
    fform : FormContainer
        Instance of :class:`~shgpy.core.data_handler.FormContainer`.

    Notes
    -----
    This routine differs from :func:`~shgpy.formgen.formgen_just_dipole_real` 
    only in the sense that the computed intensity function is computed by
    computing the modulus-squared (rather than ``**2``) of the polarization.
    For this reason, it is usually suggested to explicitly substitute 
    ``x -> real_x+1j*imag_x`` for each `x` in `t1`. See :func:`~shgpy.core.utilities.make_tensor_complex`
    and the tutorial.

    """
    c = sp.cos(theta)
    s = sp.sin(theta)
    kout = np.array([s, 0, -c], dtype=object)
    Fp = np.array([-c, 0, s], dtype=object)
    Fs = np.array([0, 1, 0], dtype=object)
    R = np.array([[sp.cos(S.phi), -sp.sin(S.phi), 0], [sp.sin(S.phi), sp.cos(S.phi), 0], [0, 0, 1]])
    rotated_tensor = tensor_contract(tensor_product(R, R, R, t1), [[1, 6], [3, 7], [5, 8]])
    Ps = tensor_contract(tensor_product(rotated_tensor, Fs, Fs), [[1, 3], [2, 4]])
    Pp = tensor_contract(tensor_product(rotated_tensor, Fp, Fp), [[1, 3], [2, 4]])
    Ps -= np.dot(kout, Ps)*kout
    Pp -= np.dot(kout, Pp)*kout
    PP = Pp[0]*sp.conjugate(Pp[0])+Pp[2]*sp.conjugate(Pp[2])
    PS = Pp[1]*sp.conjugate(Pp[1])
    SP = Ps[0]*sp.conjugate(Ps[0])+Ps[2]*sp.conjugate(Ps[2])
    SS = Ps[1]*sp.conjugate(Ps[1])
    return FormContainer({'PP':PP, 'PS':PS, 'SP':SP, 'SS':SS})


def formgen_just_dipole_real(t1, theta):
    """Generate formula assuming dipole SHG with real coefficients.

    Parameters
    ----------
    t1 : ndarray of sympy.Expr
        SHG susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    theta : float or sympy.Symbol
        Angle of incidence

    Returns
    -------
    fform : FormContainer
        Instance of :class:`~shgpy.core.data_handler.FormContainer`.

    Notes
    -----
    This routine differs from :func:`~shgpy.formgen.formgen_just_dipole_complex` 
    only in the sense that the computed intensity function is computed by
    computing ``**2`` (rather than the modulus squared) of the polarization.
    This is usually easier as long as you're not worrying about the imaginary
    component of the susceptibility (valid away from resonance in the presence
    of time-reversal symmetry).
    

    """
    c = sp.cos(theta)
    s = sp.sin(theta)
    kout = np.array([s, 0, -c], dtype=object)
    Fp = np.array([-c, 0, s], dtype=object)
    Fs = np.array([0, 1, 0], dtype=object)
    R = np.array([[sp.cos(S.phi), -sp.sin(S.phi), 0], [sp.sin(S.phi), sp.cos(S.phi), 0], [0, 0, 1]])
    rotated_tensor = tensor_contract(tensor_product(R, R, R, t1), [[1, 6], [3, 7], [5, 8]])
    Ps = tensor_contract(tensor_product(rotated_tensor, Fs, Fs), [[1, 3], [2, 4]])
    Pp = tensor_contract(tensor_product(rotated_tensor, Fp, Fp), [[1, 3], [2, 4]])
    Ps -= np.dot(kout, Ps)*kout
    Pp -= np.dot(kout, Pp)*kout
    PP = Pp[0]**2+Pp[2]**2
    PS = Pp[1]**2
    SP = Ps[0]**2+Ps[2]**2
    SS = Ps[1]**2
    return FormContainer({'PP':PP, 'PS':PS, 'SP':SP, 'SS':SS})


def formgen_dipole_quadrupole_real(t1, t2, theta):
    """Generate formula assuming dipole+quadrupole SHG with real coefficients.

    Parameters
    ----------
    t1 : ndarray of sympy.Expr
        SHG dipole susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    t2 : ndarray of sympy.Expr
        SHG quadrupole susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    theta : float or sympy.Symbol
        Angle of incidence

    Returns
    -------
    fform : FormContainer
        Instance of :class:`~shgpy.core.data_handler.FormContainer`.

    Notes
    -----
    See note in :func:`~shgpy.formgen.formgen_just_dipole_real`.
    
    """
    c = sp.cos(theta)
    s = sp.sin(theta)
    kin = np.array([s, 0, c], dtype=object)
    kout = np.array([s, 0, -c], dtype=object)
    Fp = np.array([-c, 0, s], dtype=object)
    Fs = np.array([0, 1, 0], dtype=object)
    R = np.array([[sp.cos(S.phi), -sp.sin(S.phi), 0], [sp.sin(S.phi), sp.cos(S.phi), 0], [0, 0, 1]])
    rotated_tensor = tensor_contract(tensor_product(R, R, R, t1), [[1, 6], [3, 7], [5, 8]])
    rotated_qtensor = tensor_contract(tensor_product(R, R, R, R, t2), [[1, 8], [3, 9], [5, 10], [7, 11]])
    Ps = tensor_contract(tensor_product(rotated_tensor, Fs, Fs), [[1, 3], [2, 4]])
    Pp = tensor_contract(tensor_product(rotated_tensor, Fp, Fp), [[1, 3], [2, 4]])
    Qs = tensor_contract(tensor_product(rotated_qtensor, kin, Fs, Fs), [[1, 4], [2, 5], [3, 6]])
    Qp = tensor_contract(tensor_product(rotated_qtensor, kin, Fp, Fp), [[1, 4], [2, 5], [3, 6]])
    Ps -= np.dot(kout, Ps)*kout
    Pp -= np.dot(kout, Pp)*kout
    Qs -= np.dot(kout, Qs)*kout
    Qp -= np.dot(kout, Qp)*kout
    PP = Pp[0]**2+Pp[2]**2+Qp[0]**2+Qp[2]**2
    PS = Pp[1]**2+Qp[1]**2
    SP = Ps[0]**2+Ps[2]**2+Qs[0]**2+Qs[2]**2
    SS = Ps[1]**2+Qs[1]**2
    return FormContainer({'PP':PP, 'PS':PS, 'SP':SP, 'SS':SS})


def formgen_dipole_quadrupole_complex(t1, t2, theta):
    """Generate formula assuming dipole+quadrupole SHG with complex coefficients.

    Parameters
    ----------
    t1 : ndarray of sympy.Expr
        SHG dipole susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    t2 : ndarray of sympy.Expr
        SHG quadrupole susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    theta : float or sympy.Symbol
        Angle of incidence

    Returns
    -------
    fform : FormContainer
        Instance of :class:`~shgpy.core.data_handler.FormContainer`.

    Notes
    -----
    See note in :func:`~shgpy.formgen.formgen_just_dipole_complex`.
    
    """
    theta = sp.sympify(theta)
    if not theta.is_real:
        raise ValueError('theta must be a real variable (did you forget real=True in sympy.symbols?)')
    c = sp.cos(theta)
    s = sp.sin(theta)
    kin = np.array([s, 0, c], dtype=object)
    kout = np.array([s, 0, -c], dtype=object)
    Fp = np.array([-c, 0, s], dtype=object)
    Fs = np.array([0, 1, 0], dtype=object)
    R = np.array([[sp.cos(S.phi), -sp.sin(S.phi), 0], [sp.sin(S.phi), sp.cos(S.phi), 0], [0, 0, 1]])
    rotated_tensor = tensor_contract(tensor_product(R, R, R, t1), [[1, 6], [3, 7], [5, 8]])
    rotated_qtensor = tensor_contract(tensor_product(R, R, R, R, t2), [[1, 8], [3, 9], [5, 10], [7, 11]])
    Ps = tensor_contract(tensor_product(rotated_tensor, Fs, Fs), [[1, 3], [2, 4]])
    Pp = tensor_contract(tensor_product(rotated_tensor, Fp, Fp), [[1, 3], [2, 4]])
    Qs = tensor_contract(tensor_product(rotated_qtensor, kin, Fs, Fs), [[1, 4], [2, 5], [3, 6]])
    Qp = tensor_contract(tensor_product(rotated_qtensor, kin, Fp, Fp), [[1, 4], [2, 5], [3, 6]])
    Ps -= np.dot(kout, Ps)*kout
    Pp -= np.dot(kout, Pp)*kout
    Qs -= np.dot(kout, Qs)*kout
    Qp -= np.dot(kout, Qp)*kout
    Sp = Pp+1j*Qp
    Ss = Ps+1j*Qs
    PP = Sp[0]*sp.conjugate(Sp[0])+Sp[2]*sp.conjugate(Sp[2])
    PS = Sp[1]*sp.conjugate(Sp[1])
    SP = Ss[0]*sp.conjugate(Ss[0])+Ss[2]*sp.conjugate(Ss[2])
    SS = Ss[1]*sp.conjugate(Ss[1])
    return FormContainer({'PP':PP, 'PS':PS, 'SP':SP, 'SS':SS})
